from dataclasses import dataclass
from typing import List, Dict, Callable, Any, Type

from sqlalchemy import inspect, Table
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import Session, Query, Mapper

import graphene
from graphene import ID, Argument
from graphene_sqlalchemy.converter import convert_sqlalchemy_composite
from graphene_sqlalchemy.registry import get_global_registry

from sqlalchemy_graphql_schemagen.baseclasses.logging import SimpleLoggableBase

from sqlalchemy_graphql_schemagen.graphql.schemagen.extra import create_or_get_graphql_filter_op_type_class
from sqlalchemy_graphql_schemagen.graphql.schemagen.hooks import SchemaGenHooksBase, HookDictType
from sqlalchemy_graphql_schemagen.graphql.schemagen.utilities import (
    get_graphql_field_type_for_sa_column,
    mask_ID_to_Int,
    get_columns_from_sa_model_class,
    sa_instance_to_dict,
    make_resolve_func_maker,
    gql_query_build_sa_obj_type,
    make_gql_object_description_from_sa_model_class,
    get_all_sa_model_classes,
    create_gql_update_input_object_type_from_sa_class,
    create_gql_mutation_update_arguments_class,
    create_gql_get_query_filter_input_object_type_from_sa_model,
    create_gql_get_query_order_by_filter_input_object_type_from_sa_model,
    create_create_obj_mutation_object,
    create_delete_obj_mutation_object,
    create_update_obj_mutation_object,
    get_all_sa_classless_tables,
    is_association_table)


class SQLAlchemyGraphQLSchemaGenerator(SimpleLoggableBase):
    def __init__(self, api_name: str, declarative_base: DeclarativeMeta, sa_connection_string: str,
                 ignore_models: List = None, op_hooks: HookDictType = None, sa_composite_converters: Dict[Any, Any] = None,
                 extra_query_objects: Dict[str, graphene.ObjectType] = None,
                 extra_mutation_objects: Dict[str, graphene.ObjectType] = None,
                 extra_object_meta_args: Dict[DeclarativeMeta, Dict[str, object]] = None):

        # List of Models to ignore during dynamic SQLAlchemy-->GraphQL generation

        self.ignore_models = ignore_models if ignore_models else []

        # Extra (query|mutation|object meta) objects
        self.extra_mutation_objects = extra_mutation_objects if extra_mutation_objects else {}
        self.extra_query_objects = extra_query_objects if extra_query_objects else {}
        self.extra_object_meta_args = extra_object_meta_args if extra_object_meta_args else {}

        self.api_name = api_name
        self.declarative_base = declarative_base

        # save the connection string
        self.sa_connection_string = sa_connection_string

        # save the hooks
        self.op_hooks = op_hooks or {}

        # Save and register SQLAlchemy Composite Types
        self.sa_composite_converters = sa_composite_converters or {}
        self.register_composites()

    def register_composites(self):
        for (
            CompositeClass,
            CompositeConverterFunc,
        ) in self.sa_composite_converters.items():
            # Generate a Default Converter (STRING)
            def default_sa_composite_class_converter(composite, _registry):
                return graphene.String(description=composite.doc)

            # Register converter using the provided function or the default one will be provided.
            convert_sqlalchemy_composite.register(CompositeClass)(
                CompositeConverterFunc
                if callable(CompositeConverterFunc)
                else default_sa_composite_class_converter
            )

    ##################################################
    # Main Function - Entry Point
    ##################################################
    def get_graphene_schema(
        self, generate_query=True, generate_mutation=True, **graphene_schema_args
    ) -> graphene.Schema:

        schema_params = {"query": None, "mutation": None}
        schema_params.update(graphene_schema_args)

        if generate_query:
            self.l.debug("Generating Query schema...")
            schema_params["query"] = self.generate_query_schema()

        if generate_mutation:
            self.l.debug("Generating Mutation schema...")
            schema_params["mutation"] = self.generate_mutation_schema()

        self.l.debug("Assembling graphene.Schema from schema_params")

        schema = graphene.Schema(**schema_params)
        self.l.debug(
            f"""
---GRAPHQL SCHEMA START---
{str(schema)}
---GRAPHQL SCHEMA END---
"""
        )

        return graphene.Schema(**schema_params)

    ##################################################
    # Generate QUERY Schema
    ##################################################
    def generate_query_schema(self) -> type:

        # List of all Model Classes from SQLAlchemy
        sa_model_classes = get_all_sa_model_classes(self.declarative_base)

        # Root Query Class Attribute Dict
        root_query_class_dict = {"__doc__": f'Root Query Class for "{self.api_name}"'}

        # List of All Object to Generate Dynamic Methods
        all_objects = sa_model_classes

        # Iterate through all SQLAlchemy's classes, building GraphQL Objects
        for sa_queryable_object in all_objects:

            self.generate_query_schema_sa_class(
                root_query_class_dict, sa_queryable_object
            )

        # Iterate through extra_query_objects
        for obj_name, graphene_obj in self.extra_query_objects.items():
            self.l.debug(f"Attaching Extra_Query_Object {obj_name}...")
            root_query_class_dict[obj_name] = graphene_obj

        # Build Main Class and attach objects built in the last step.
        root_query_class = type(
            f"Query_{self.api_name}", (graphene.ObjectType,), root_query_class_dict
        )

        # Return the newly built class.
        return root_query_class

    def generate_query_schema_sa_table(
        self, root_query_class_dict, sa_queryable_object: Table
    ):
        sa_table_name = str(sa_queryable_object.name)
        self.l.debug(f"[Query] SQLAlchemy Classless Table --> {sa_table_name}")

        # Get resolve_<object> Function
        resolve_func = make_resolve_func_maker(
            sa_queryable_object, self.get_session_func, self.op_hooks
        )

        # Attach resolve func to class.
        root_query_class_dict[resolve_func.__name__] = resolve_func
        # Build {Class} = graphene.List({Class})
        graphql_model_class = gql_query_build_sa_obj_type(sa_queryable_object)
        # Attach {Class} = graphene.List({Class})
        root_query_class_dict[f"{sa_table_name}"] = graphene.List(
            graphql_model_class,
            description=make_gql_object_description_from_sa_model_class(
                sa_queryable_object
            ),
            filters=create_gql_get_query_filter_input_object_type_from_sa_model(
                sa_queryable_object
            ),
            order_by=create_gql_get_query_order_by_filter_input_object_type_from_sa_model(
                sa_queryable_object
            ),
            page=Argument(
                graphene.Int, default_value=1, description="(Pagination) Page Number"
            ),
            perpage=Argument(
                graphene.Int,
                default_value=50,
                description="(Pagination) Results Per Page",
            ),
        )

    def generate_query_schema_sa_class(
        self, root_query_class_dict, sa_queryable_object
    ):
        # Get the actual table name for this entity.
        sa_table_name = sa_queryable_object.__tablename__
        self.l.debug(f"[Query] SQLAlchemy Class --> {sa_queryable_object.__name__}")
        # Get resolve_<object> Function
        resolve_func = make_resolve_func_maker(
            sa_queryable_object, self.sa_connection_string, self.op_hooks
        )

        extra_metaclass_properties = {}
        if sa_queryable_object in self.extra_object_meta_args:
            extra_metaclass_properties = self.extra_object_meta_args[sa_queryable_object]

        # Build {Class} = graphene.ObjectType({Class})
        graphql_model_class = gql_query_build_sa_obj_type(sa_queryable_object,
                                                          extra_metaclass_properties=extra_metaclass_properties)

        if sa_queryable_object not in self.ignore_models:
            # Attach resolve func to class.
            root_query_class_dict[resolve_func.__name__] = resolve_func

            # Attach {Class} = graphene.List({Class})
            root_query_class_dict[f"{sa_table_name}"] = graphene.List(
                graphql_model_class,
                description=make_gql_object_description_from_sa_model_class(
                    sa_queryable_object
                ),
                filters=create_gql_get_query_filter_input_object_type_from_sa_model(
                    sa_queryable_object
                ),
                order_by=create_gql_get_query_order_by_filter_input_object_type_from_sa_model(
                    sa_queryable_object
                ),
                page=Argument(
                    graphene.Int, default_value=1, description="(Pagination) Page Number"
                ),
                perpage=Argument(
                    graphene.Int,
                    default_value=50,
                    description="(Pagination) Results Per Page",
                ),
            )


    ##################################################
    # Generate Mutation Schema
    ##################################################
    def generate_mutation_schema(self) -> type:

        root_mutation_class_dict = {
            "__doc__": f'Root Mutation Class for "{self.api_name}"'
        }
        for sa_model_class in get_all_sa_model_classes(self.declarative_base):
            if sa_model_class in self.ignore_models:
                continue

            class_name = sa_model_class.__name__
            self.l.debug(f"[Mutation] SQLAlchemy Class --> {class_name}")

            ################################
            # UPDATE - skip on many-to-many association table
            ################################
            if not is_association_table(sa_model_class):
                update_obj_class = create_update_obj_mutation_object(
                    sa_model_class, self.sa_connection_string, self.op_hooks
                )

                root_mutation_class_dict[
                    f"update_{class_name.lower()}"
                ] = update_obj_class.Field()
            else:
                self.l.debug(
                    f'Skipping "UPDATE" on class "{class_name}" since it is a many-to-many assoc table...'
                )

            ################################
            # CREATE
            ################################
            create_obj_class = create_create_obj_mutation_object(
                sa_model_class, self.sa_connection_string, self.op_hooks
            )

            root_mutation_class_dict[
                f"create_{class_name.lower()}"
            ] = create_obj_class.Field()

            ################################
            # DELETE
            ################################
            delete_obj_class = create_delete_obj_mutation_object(
                sa_model_class, self.sa_connection_string, self.op_hooks
            )

            root_mutation_class_dict[
                f"delete_{class_name.lower()}"
            ] = delete_obj_class.Field()


        # Iterate through extra_mutation_objects
        for obj_name, graphene_obj in self.extra_mutation_objects.items():
            self.l.debug(f"Attaching Extra_Mutation_Object {obj_name}...")
            root_mutation_class_dict[obj_name] = graphene_obj

        root_mutation_class = type(
            f"Mutation_{self.api_name}",
            (graphene.ObjectType,),
            root_mutation_class_dict,
        )

        return root_mutation_class



from decimal import Decimal


class Money(object):
    """Class to Represent a Monetary Value"""

    def __init__(self, currency: str, value: Decimal):
        self.currency = currency
        self.value = value

    def __composite_values__(self):
        return self.currency, self.value

    def __repr__(self):
        return "MoneyComposite(curr=%r, value=%r)" % (self.currency, self.value)

    def __eq__(self, other):
        return (
            isinstance(other, Money)
            and other.currency == self.currency
            and other.value == self.value
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return f"{self.currency if self.currency else ''} {self.value if self.value else ''}".strip()



import logging


# Class that can be inherited to provide a .l logger method.
class SimpleLoggableBase(object):
    @property
    def l(self) -> logging.Logger:

        # Get RootLogger
        rootlogger = logging.getLogger()

        # Try to get this class' logger
        loggerobj = getattr(self, "__logger__", None)

        if not loggerobj:
            # Create this class logger if it doesn't exist.
            self.__logger__ = rootlogger.getChild(self.__class__.__name__)

        # return the logger
        return self.__logger__



import graphene
from graphene.utils.subclass_with_meta import SubclassWithMeta_Meta


################################
# Order By Operations
################################
class OrderByOperation(graphene.Enum):
    ASC = "ASC"
    DESC = "DESC"


################################
# Filter Operations
################################
class FilterOperation(graphene.Enum):
    EQ = "EQ"
    NEQ = "NEQ"

    IS = "IS"
    ISNOT = "ISNOT"
    ISNULL = "ISNULL"
    ISNOTNULL = "ISNOTNULL"

    LT = "LT"
    GT = "GT"

    LIKE = "LIKE"
    NOTLIKE = "NOTLIKE"

    ILIKE = "ILIKE"
    NOTILIKE = "NOTILIKE"

    # List operator
    IN = "IN"
    NOTIN = "NOTIN"
    BETWEEN = "BETWEEN"




################################
# <type>FilterOp GraphQL Object
################################

__SCHEMAGEN_filter_op_type_class_registry = {}


def create_or_get_graphql_filter_op_type_class(graphql_type: SubclassWithMeta_Meta):
    # Global repository of <typed>FilterOperations
    global __SCHEMAGEN_filter_op_type_class_registry

    # The GraphQL type name we are going to create a filter op type to.
    graphql_type_name = graphql_type.__name__

    if graphql_type_name not in __SCHEMAGEN_filter_op_type_class_registry:
        # Create
        fop_class_name = f"{graphql_type_name}FilterOp"
        fop = type(
            fop_class_name,
            (graphene.InputObjectType,),
            {
                "op": graphene.Field(FilterOperation, required=True),
                "v": graphene.Field(graphql_type, required=False, default_value=None),
                "vl": graphene.Field(graphene.List(graphql_type), required=False, default_value=None)
            },
        )

        __SCHEMAGEN_filter_op_type_class_registry[graphql_type_name] = fop

    return __SCHEMAGEN_filter_op_type_class_registry[graphql_type_name]



from enum import Enum
from typing import Dict, Type, Any

from abc import ABC, abstractmethod

from graphql import ResolveInfo


class HookOperation(Enum):
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    DELETE = "delete"


class PreHookResult(Enum):
    CONTINUE = 1
    STOP = 2


class SchemaGenHookException(Exception):
    pass


# noinspection PyMethodParameters
class SchemaGenHooksBase(ABC):
    @staticmethod
    @abstractmethod
    def pre(parent, info: ResolveInfo, **kwargs) -> PreHookResult:
        """
        root = root value from Schema
        info = gathers additional request oriented-stuff, like info.context
        **kwargs = whatever else would be sent to the resolve/mutate function.
        """
        return PreHookResult.CONTINUE

    @staticmethod
    @abstractmethod
    def post(parent, info: ResolveInfo, func_retval, **kwargs) -> Any:
        """
        root = root value from Schema
        info = gathers additional request oriented-stuff, like info.context
        **kwargs = whatever else would be sent to the resolve/mutate function.
        kwargs['__func_retval'] contains the decorated function's return value, so you can postprocess it.

        OBS: anything that gets returned here will substitute the decorated function's return value
        """
        pass

    def __init__(self, func_to_be_decorated):
        self.func_to_be_decorated = func_to_be_decorated

    def __call__(self, *args, **kwargs):
        # Call pre-hook
        pre_result = self.pre(*args, **kwargs)
        if pre_result == PreHookResult.STOP:
            raise SchemaGenHookException("Pre_Hook called STOP")

        # Call Function
        decorated_func_retval = self.func_to_be_decorated(*args, **kwargs)

        # Call post-hook with the original arguments and the decorated function's returned value.
        post_retval = self.post(*args, decorated_func_retval, **kwargs)
        if post_retval:
            # post wants to substitute the original function's data with its own.
            return post_retval

        # No 'post' hook data, return the original function's data, unmodified.
        return decorated_func_retval


# Custom Type
HookDictType = Dict[HookOperation, Type[SchemaGenHooksBase]]



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="sqlalchemy-graphql-schemagen",
    license="MIT",
    version="1.0.6",
    author="flipbit03",
    author_email="cadu.coelho@gmail.com",
    description="Generate a full (query+mutation) GraphQL schema from your SQLAlchemy Declarative Model Base.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flipbit03/sqlalchemy-graphql-schemagen",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "cryptography>=3.3.2",
        "flask>=1.1.1",
        "flask-graphql>=2.0.1",
        "graphene-sqlalchemy>=2.2",
        "sqlalchemy>=1.3.13",
    ],
)



import gc
from contextlib import contextmanager
from typing import List, Dict, Callable, Union

import graphene
import sqlalchemy
from graphene.utils.str_converters import to_camel_case
from graphene.utils.subclass_with_meta import SubclassWithMeta_Meta
from graphene_sqlalchemy import SQLAlchemyObjectType
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
from graphene_sqlalchemy.registry import get_global_registry
from graphql import GraphQLError
from sqlalchemy import Column, inspect, ColumnDefault, Table, create_engine
from sqlalchemy.exc import IntegrityError, DBAPIError
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import Mapper, Session, Query
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.state import InstanceState
from sqlalchemy.pool import NullPool

from . import HookDictType
from .extra import (
    FilterOperation,
    create_or_get_graphql_filter_op_type_class,
    OrderByOperation,
)

# this variable holds a list of all graphene.Enums created
# with the table field names.
#
# we need to cache this and reuse accordingly since you cannot
# have repeated schema names in a GraphQL Schema definition.
from .hooks import HookOperation

################################
################################
# Global Vars and Registries
################################
################################

__SCHEMAGEN_field_enum_by_sa_type_registry = {}


################################
# Custom SQLAlchemyObjectType Class - so we can define shared properties.
################################
class OurBaseSQLAlchemyObjectType(SQLAlchemyObjectType):
    class Meta:
        abstract = True


################################
################################
# SchemaGen's GraphQL Class Builder Helpers
################################
################################

################################
# Function to check if a SQLAlchemy's Model Class is Associative
################################
def is_association_table(sa_model_class) -> bool:
    mci: Mapper = inspect(sa_model_class)
    return list(mci.columns.values()) == list(mci.primary_key)


################################
# Get the SqlAlchemy's "Queryable Object"'s name (be it a Class or Table)
################################
def get_sa_queryable_name(sa_queryable_obj):
    if isinstance(sa_queryable_obj, Table):
        table_name = str(sa_queryable_obj.name)
    else:
        mci: Mapper = inspect(sa_queryable_obj)
        table_name = str(mci.tables[0].name)
    return table_name


################################
# Get all SQLAlchemy's Model Classes as a List
################################
def get_all_sa_model_classes(
        sa_model_base_class: DeclarativeMeta,
) -> List[DeclarativeMeta]:
    # Get SQLAlchemy Class Registry
    sa_class_registry: dict = getattr(sa_model_base_class, "_decl_class_registry")

    # List of all Model Classes from SQLAlchemy
    model_classes = [x for x in sa_class_registry.values() if hasattr(x, "__table__")]

    return model_classes


################################
# Get all SQLAlchemy's Model Classless Tables (ex:many2many) as a List
################################
def get_all_sa_classless_tables(sa_model_base_class: DeclarativeMeta, ) -> List[Table]:
    # Get SQLAlchemy Class Registry
    sa_class_registry: dict = getattr(sa_model_base_class, "_decl_class_registry")

    # List of all Model Classes from SQLAlchemy
    model_classes = get_all_sa_model_classes(sa_model_base_class)
    model_classes_table_list = [o.__table__ for o in model_classes]

    # All Tables
    all_tables = sa_model_base_class.metadata.tables.values()

    # Extract Remaining tables.
    classless_sa_tables = [
        table_obj
        for table_obj in all_tables
        if table_obj not in model_classes_table_list
    ]

    return classless_sa_tables


##################################################
# SQLAlchemySchemaGenerator Helper Functions
##################################################
def create_input_field_args(sa_column) -> dict:
    field_args = {}

    ################################
    # Field Docstring
    ################################
    # Field doc= argument
    sa_column_doc_string = getattr(sa_column, "doc", None)
    if sa_column_doc_string:
        field_args["description"] = sa_column_doc_string

    # If the field is a primary_key, make it OPTIONAL
    if sa_column.primary_key:
        ################################
        # Primary Key - Make it Optional and Return
        ################################
        field_args["required"] = False
    else:
        ################################
        # Normal Column
        ################################

        # Is this column NOT NULL? make it required
        if not sa_column.nullable:
            field_args["required"] = True

        # do we have a default value on this column?
        if sa_column.default:
            cd: ColumnDefault = sa_column.default

            # Disable 'required' above
            field_args["required"] = False

            # is the default an scalar
            if cd.is_scalar:
                # set the default value from that scalar.
                field_args["default_value"] = cd.arg

    return field_args


def gql_query_build_sa_obj_type(sa_queryable_object: Union[DeclarativeMeta, Table],
                                extra_metaclass_properties: Dict[str, object] = None):
    if not extra_metaclass_properties:
        extra_metaclass_properties = {}

    # This dict holds the properties that need to exist in the metaclass
    metaclass_properties = {"model": sa_queryable_object, "description": sa_queryable_object.__doc__}

    # Set extra properties
    # Good examples:
    #   only_fields = ("name",)
    #   exclude_fields = ("last_name",)
    metaclass_properties.update(extra_metaclass_properties)

    meta_model_class = type(
        "Meta",
        (),
        metaclass_properties,
    )

    sa_obj_type_class = type(
        f"{get_sa_queryable_name(sa_queryable_object)}",
        (OurBaseSQLAlchemyObjectType,),
        {"Meta": meta_model_class},
    )
    return sa_obj_type_class


################################
# SQLAlchemy Operations for order_by
################################
order_by_ops = {"ASC": sqlalchemy.asc, "DESC": sqlalchemy.desc}


################################
# Create the resolve_MODELs function for "get_all"
# - with filter
# - with pagination
################################
def make_resolve_func_maker_func_name(sa_queryable_obj) -> str:
    # Generate Resolver Function Name
    # Example: "resolve_Users"

    queryable_name = get_sa_queryable_name(sa_queryable_obj)
    funcname = f"resolve_{queryable_name}"

    return funcname


################################
# create an independent, scoped db connection
# with a well defined lifetime
# uses 'context manager' so we can use the with: protocol
################################
@contextmanager
def scoped_db_session_from_sa_connection_string(sa_connection_string: str) -> Session:
    """ Creates a context with an open SQLAlchemy session.
    """
    engine = create_engine(sa_connection_string, convert_unicode=True, poolclass=NullPool)
    connection = engine.connect()
    scoped_db_session = scoped_session(
        sessionmaker(autocommit=False, autoflush=True, bind=engine)
    )
    yield scoped_db_session

    # Leave the session and connection open, GraphQL
    # scoped_db_session.close()
    # connection.close()


def make_resolve_func_maker(
        sa_queryable_obj: DeclarativeMeta, sa_connection_string: str, hooks: HookDictType
) -> Callable:
    # Get Mapper for SQLAlchemy's Model Class

    m: Mapper = inspect(sa_queryable_obj)

    def resolve_func(_parent, _info, **kwargs):
        with scoped_db_session_from_sa_connection_string(sa_connection_string) as s:

            nonlocal sa_queryable_obj

            ################################
            # Base Query
            ################################
            q = s.query(sa_queryable_obj)

            ################################
            # Filter
            ################################

            # List of Filters
            filter_obj_list = kwargs.get("filters", {})

            for filter_obj in filter_obj_list:
                for filter_name in filter_obj:
                    filter_obj = filter_obj[filter_name]

                    sa_column: Column = m.columns[filter_name]

                    if FilterOperation.get(filter_obj.op) == FilterOperation.EQ:
                        q = q.filter(sa_column == filter_obj.v)
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.NEQ:
                        q = q.filter(sa_column != filter_obj.v)
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.IS:
                        q = q.filter(sa_column.is_(filter_obj.v))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.ISNOT:
                        q = q.filter(sa_column.isnot(filter_obj.v))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.ISNULL:
                        q = q.filter(sa_column.is_(None))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.ISNOTNULL:
                        q = q.filter(sa_column.isnot(None))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.LT:
                        q = q.filter(sa_column < filter_obj.v)
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.GT:
                        q = q.filter(sa_column > filter_obj.v)
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.LIKE:
                        q = q.filter(sa_column.like(f"%{filter_obj.v}%"))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.NOTLIKE:
                        q = q.filter(sa_column.notlike(f"%{filter_obj.v}%"))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.ILIKE:
                        q = q.filter(sa_column.ilike(f"%{filter_obj.v}%"))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.NOTILIKE:
                        q = q.filter(sa_column.notilike(f"%{filter_obj.v}%"))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.IN:
                        q = q.filter(sa_column.in_(filter_obj.vl))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.NOTIN:
                        q = q.filter(sa_column.notin_(filter_obj.vl))
                    elif FilterOperation.get(filter_obj.op) == FilterOperation.BETWEEN:
                        q = q.filter(sa_column.between(filter_obj.vl[0], filter_obj.vl[1]))

            ################################
            # ORDER_BY
            ################################
            order_by_params = kwargs.get("order_by")

            if order_by_params:
                order_by_function = order_by_ops[order_by_params.o]
                order_column = m.columns[order_by_params.f]

                q = q.order_by(order_by_function(order_column))

            ################################
            # LIMIT / OFFSET (Pagination)
            ################################

            # Pagination
            page = max(0, kwargs.get("page") - 1)
            perpage = kwargs.get("perpage")

            # Apply Pagination
            q = q.limit(perpage).offset(page * perpage)

            results = q.all()

            #gc.collect()

        return results

    # Bind final_resolve_func to the vanilla (without hooks) resolve function.
    final_resolve_func = resolve_func

    # Decorate resolve_func with the provided hooks class, if we have one.
    read_hook = hooks.get(HookOperation.READ)
    if read_hook:
        final_resolve_func = read_hook(resolve_func)

    # Generate Resolver Function Name
    # Example: "resolve_Users"
    final_resolve_func_name = make_resolve_func_maker_func_name(sa_queryable_obj)
    final_resolve_func.__name__ = final_resolve_func_name

    return final_resolve_func


################################
# Generate Documentation for each SQLAlchemy Model that is going to be mapped to a GraphQL Type
################################
def make_gql_object_description_from_sa_model_class(mc: DeclarativeMeta) -> str:
    mci: Mapper = inspect(mc)

    # Generate attributes to apply inside the docs.
    class_docstring = f'"{mc.__doc__}"\\\n' if mc.__doc__ else ""
    primary_key_name: str = mci.primary_key[0].key
    return f"""{class_docstring}pk: \"{primary_key_name}\""""


################################
# Convert any SQLAlchemy Model Instance (Query Result) to a plain dict()
################################
def sa_instance_to_dict(mc: DeclarativeMeta):
    """Transform a SQLAlchemy instance to a sanitized dict, without instance/control vars"""
    mci: InstanceState = inspect(mc)

    column_list = get_columns_from_sa_model_class(mci.class_)

    # get obj's columns from <Mapper>.columns and do a direct getattr() using those names
    return {c.name: getattr(mc, c.name) for c in column_list}


################################
# Grab a List of SQLAlchemy Columns from a SQLAlchemy Model
################################
def get_columns_from_sa_model_class(sa_model_class: DeclarativeMeta) -> List[Column]:
    # Model Classes return a inspection-able SQLAlchemy "Mapper" object.
    mci: Mapper = inspect(sa_model_class)

    return [column for column in mci.columns]


################################
# Grab GraphQL Field from a SQLAlchemy Column
################################

mask_type = Dict[SubclassWithMeta_Meta, SubclassWithMeta_Meta]

# by default, remap ID to "Int" for filter operations
mask_ID_to_Int: mask_type = {graphene.ID: graphene.Int}


def get_graphql_field_type_for_sa_column(
        column: Column, mask: mask_type = None
) -> SubclassWithMeta_Meta:
    original_graphql_type = convert_sqlalchemy_type(
        column.type, column, get_global_registry()
    )

    # Mask type using the mask argument.
    if mask:
        returned_graphql_type = mask.get(original_graphql_type, original_graphql_type)
        return returned_graphql_type

    # Normal return
    return original_graphql_type


################################
################################
# READ / QUERY
################################
################################


def create_gql_get_query_filter_input_object_type_from_sa_model(mc: DeclarativeMeta):
    sa_column_list = get_columns_from_sa_model_class(mc)

    class_items = {}
    for sa_column in sa_column_list:
        # For filter definitions, use 'Int' as type for 'Id'
        graphql_column_type = get_graphql_field_type_for_sa_column(
            sa_column, mask=mask_ID_to_Int
        )

        fop_class = create_or_get_graphql_filter_op_type_class(graphql_column_type)

        class_items[sa_column.name] = graphene.Field(fop_class)

    query_filter_params_input_object_type = type(
        f"{mc.__name__}QueryParams", (graphene.InputObjectType,), class_items
    )

    # Remember to wrap this around an graphene.Argument so it can be used in an actual field parameter
    return graphene.Argument(graphene.List(query_filter_params_input_object_type))


################################
# Create Graphene.Enum with the Column Names from a SQLAlchemy Model Class
################################
def create_or_get_gql_field_list_enum_from_sa_model(mc: DeclarativeMeta):
    global __SCHEMAGEN_field_enum_by_sa_type_registry
    field_enum_registry = __SCHEMAGEN_field_enum_by_sa_type_registry

    mc_name = mc.__name__
    if mc_name in field_enum_registry:
        return field_enum_registry[mc_name]

    # Todo: Respect camelCase setting from graphene. for now this is hardcoded to convert to camelcase
    def gen_enum_entry(field_name):
        return to_camel_case(field_name), field_name

    sa_column_list = get_columns_from_sa_model_class(mc)
    return graphene.Enum(
        f"{mc.__name__}FieldEnum", [gen_enum_entry(_c.name) for _c in sa_column_list]
    )


################################
# Query ORDER_BY
################################


def create_gql_get_query_order_by_filter_input_object_type_from_sa_model(
        mc: DeclarativeMeta,
):
    field_enum = create_or_get_gql_field_list_enum_from_sa_model(mc)

    class_items = {
        "f": graphene.Argument(field_enum, required=True),
        "o": graphene.Argument(OrderByOperation, required=True),
    }

    query_filter_params_input_object_type = type(
        f"{mc.__name__}OrderByParams", (graphene.InputObjectType,), class_items
    )

    # Remember to wrap this around an graphene.Argument so it can be used in an actual field parameter
    return graphene.Argument(query_filter_params_input_object_type)


################################
################################
# UPDATE
################################
################################

################################
# Object that validates the input arguments of the Update<Model> function
################################
def create_gql_update_input_object_type_from_sa_class(
        sa_model_class: DeclarativeMeta,
) -> type:
    sa_column_list = get_columns_from_sa_model_class(sa_model_class)

    input_class_fields = {}

    for sa_column in sa_column_list:
        graphql_type = get_graphql_field_type_for_sa_column(sa_column)

        field_args = {}
        # is this the Primary_key?
        if graphql_type == graphene.ID:
            field_args["required"] = True

        # Field doc= argument
        sa_column_doc_string = getattr(sa_column, "doc", None)
        if sa_column_doc_string:
            field_args["description"] = sa_column_doc_string

        # Populate Class Dict
        input_class_fields[sa_column.name] = graphql_type(**field_args)

    input_class_name = f"{sa_model_class.__name__}UpdateInput"
    input_class = type(
        input_class_name, (graphene.InputObjectType,), input_class_fields
    )

    return input_class


################################
# "Update" Mutation's Argument Class
# use the method above to get the proper input object type for this
################################
def create_gql_mutation_update_arguments_class(sa_model_class: DeclarativeMeta) -> type:
    obj_input_class = create_gql_update_input_object_type_from_sa_class(sa_model_class)

    # This is where we'll populate the soon-to-be updated class' arguments
    argument_name = f"{sa_model_class.__name__.lower()}_data"

    arg_class_items = {argument_name: obj_input_class(required=True)}

    arguments_class = type("Arguments", (), arg_class_items)

    return arguments_class


################################
# Get the Original Object by PRIMARY KEY ID
################################
def get_sa_obj_by_pk_id(
        sa_class: DeclarativeMeta, pk_name: str, pk_id: int, s: Session
) -> DeclarativeMeta:
    x = s.query(sa_class).filter(getattr(sa_class, pk_name) == int(pk_id)).one()
    return x


################################
# update<Model> MAIN FUNCTION
################################
def create_update_obj_mutation_object(
        sa_model_class: DeclarativeMeta, sa_connection_string: str, hooks: HookDictType
) -> type:
    cls_name: str = sa_model_class.__name__

    # use the same name as the SQLAlchemy's class name
    updated_graphql_obj_name = cls_name

    # get the graphql object associated with this SQLAlchemy Model Class
    gql_object = get_global_registry().get_type_for_model(sa_model_class)

    ################################
    # create a partial class to be able to use it inside the mutate() function ;-)
    ################################

    update_obj_partial_class_items = {
        # Update Mutation Arguments
        "Arguments": create_gql_mutation_update_arguments_class(sa_model_class),
        # The return value of the update mutation query (the same object)
        updated_graphql_obj_name: graphene.Field(lambda: gql_object),
        # this needs to exist or else the class cannot be created, it needs this item to exist.
        "mutate": lambda: None,
    }

    # Create Partial UpdateObj Class
    update_obj_partial_class = type(
        f"PartialUpdate{cls_name}", (graphene.Mutation,), update_obj_partial_class_items
    )

    # Get class name from the outside and build the parameter name dynamically
    param_name = f"{cls_name.lower()}_data"

    # primary key name
    pk_name = inspect(sa_model_class).primary_key[0].name

    # Mutate Function Entry Point
    def mutate_func(root, info, **kwargs):

        # Get the new instance data from kwargs[param_name]
        incoming_update_request: dict = kwargs.get(param_name)

        with scoped_db_session_from_sa_connection_string(sa_connection_string) as s:
            s: Session

            # #### SQLAlchemy TIME ####

            # 1- get original entity from SA
            to_update_sa_obj = get_sa_obj_by_pk_id(
                sa_model_class, pk_name, incoming_update_request[pk_name], s
            )

            # 2- modify it
            for k, v in incoming_update_request.items():
                # skip re-setting the primary key as that can trigger unwanted dirty() pkid
                # updates that might fail.
                if k == pk_name:
                    continue

                # Update the attribute value
                setattr(to_update_sa_obj, k, v)

            try:
                # 3- add to session
                s.add(to_update_sa_obj)

                # 4- commit
                s.commit()
            except DBAPIError as e:
                s.rollback()
                s.flush()
                s.close()
                raise GraphQLError(e.args)

        partial_update_obj_class_invocation = {
            updated_graphql_obj_name: to_update_sa_obj
        }

        #gc.collect()
        return update_obj_partial_class(**partial_update_obj_class_invocation)

    # Decorate resolve_func with the provided hooks class, if we have one.
    final_mutate_func = mutate_func
    update_hook = hooks.get(HookOperation.UPDATE)
    if update_hook:
        final_mutate_func = update_hook(mutate_func)

    # Create Definitive Class
    definitive_update_obj_class_items = {"mutate": final_mutate_func}

    definitive_update_obj_class = type(
        f"Update{cls_name}",
        (update_obj_partial_class,),
        definitive_update_obj_class_items,
    )

    return definitive_update_obj_class


################################
################################
# CREATE
################################
################################


################################
# Object that validates the input arguments of the Create<Model> function
################################
def create_gql_create_input_object_type_from_sa_class(
        sa_model_class: DeclarativeMeta,
) -> type:
    sa_column_list = get_columns_from_sa_model_class(sa_model_class)

    input_class_fields = {}

    for sa_column in sa_column_list:
        graphql_type = get_graphql_field_type_for_sa_column(sa_column)

        # Generate GraphQL Field Arguments from SQLAlchemy's Column Type, Default Value, Docstring, ...
        field_args = create_input_field_args(sa_column)

        # Populate Class Dict
        input_class_fields[sa_column.name] = graphql_type(**field_args)

    input_class_name = f"{sa_model_class.__name__}CreateInput"
    input_class = type(
        input_class_name, (graphene.InputObjectType,), input_class_fields
    )

    return input_class


################################
# "Create" Mutation's Argument Class
################################
def create_gql_mutation_create_arguments_class(sa_model_class: DeclarativeMeta) -> type:
    obj_input_class = create_gql_create_input_object_type_from_sa_class(sa_model_class)

    # This is where we'll populate the soon-to-be updated class' arguments
    argument_name = f"{sa_model_class.__name__.lower()}_data"

    arg_class_items = {argument_name: obj_input_class(required=True)}

    arguments_class = type("Arguments", (), arg_class_items)

    return arguments_class


################################
# create<Model> MAIN FUNCTION
################################
def create_create_obj_mutation_object(
        sa_model_class: DeclarativeMeta, sa_connection_string, hooks: HookDictType
) -> type:
    cls_name: str = sa_model_class.__name__

    # use the same name as the SQLAlchemy's class name
    new_graphql_obj_name = cls_name

    # get the graphql object associated with this SQLAlchemy Model Class
    gql_object = get_global_registry().get_type_for_model(sa_model_class)

    ################################
    # create a partial class to be able to use it inside the mutate() function ;-)
    ################################

    create_obj_partial_class_items = {
        # Update Mutation Arguments
        "Arguments": create_gql_mutation_create_arguments_class(sa_model_class),
        # The return value of the update mutation query (the same object)
        new_graphql_obj_name: graphene.Field(lambda: gql_object),
        # this needs to exist or else the class cannot be created, it needs this item to exist.
        "mutate": lambda: None,
    }

    # Create Partial CreateObj Class
    create_obj_partial_class = type(
        f"PartialCreate{cls_name}", (graphene.Mutation,), create_obj_partial_class_items
    )

    # Get class name from the outside and build the parameter name dynamically
    param_name = f"{cls_name.lower()}_data"

    # Mutate Function Entry Point
    def mutate_func(root, info, **kwargs):
        with scoped_db_session_from_sa_connection_string(sa_connection_string) as s:
            s: Session

            # Get the new instance data from kwargs[param_name]
            create_data: dict = kwargs.get(param_name)

            ###################
            # CREATE DATA
            ###################

            # 1- Create a new object
            new_obj = sa_model_class()

            # 2- Add data from the GraphQL Request to the SQLAlchemy Object
            for k, v in create_data.items():
                setattr(new_obj, k, v)

            try:
                # 3- Add new obj to Session
                s.add(new_obj)

                # 4- Commit!
                s.commit()
            except DBAPIError as e:
                s.rollback()
                s.flush()
                s.close()
                raise GraphQLError(e.args)

            # 5- Attach the newly created object to the return object
            sa_instance_as_dict = sa_instance_to_dict(new_obj)
            partial_create_obj_class_invocation = {
                new_graphql_obj_name: new_obj
            }

        #gc.collect()

        return create_obj_partial_class(**partial_create_obj_class_invocation)

    definitive_create_function = mutate_func
    create_hook = hooks.get(HookOperation.CREATE)
    if create_hook:
        definitive_create_function = create_hook(mutate_func)

    # Create Definitive Class
    definitive_create_obj_class_items = {"mutate": definitive_create_function}

    definitive_create_obj_class = type(
        f"Create{cls_name}",
        (create_obj_partial_class,),
        definitive_create_obj_class_items,
    )

    return definitive_create_obj_class


################################
################################
# DELETE
################################
################################

################################
# "Delete" Mutation's Argument Class
################################
def create_gql_mutation_delete_arguments_class(sa_model_class: DeclarativeMeta) -> type:
    arg_class_items = {}

    primary_keys = inspect(sa_model_class).primary_key

    for pk_obj in primary_keys:
        pk_name = pk_obj.name
        arg_class_items[pk_name] = graphene.Int(required=True)

    arguments_class = type("Arguments", (), arg_class_items)

    return arguments_class


################################
# delete<Model> MAIN FUNCTION
################################
def create_delete_obj_mutation_object(
        sa_model_class: DeclarativeMeta, sa_connection_string, hooks: HookDictType
) -> type:
    # Class Name
    cls_name: str = sa_model_class.__name__

    ################################
    # create a partial class to be able to use it inside the mutate() function ;-)
    ################################

    # Argument name - the count of deleted items (1 or 0)
    deleted_count_arg_name = "deleted_count"

    delete_obj_partial_class_items = {
        # Update Mutation Arguments
        "Arguments": create_gql_mutation_delete_arguments_class(sa_model_class),
        # The return value of the update mutation query (the same object)
        deleted_count_arg_name: graphene.Field(graphene.Int),
        # this needs to exist or else the class cannot be created, it needs this item to exist.
        "mutate": lambda: None,
    }

    # Create Partial CreateObj Class
    delete_obj_partial_class = type(
        f"PartialDelete{cls_name}", (graphene.Mutation,), delete_obj_partial_class_items
    )

    # primary key name
    pk_names = [x.name for x in inspect(sa_model_class).primary_key]

    # Mutate Function Entry Point
    def mutate_func(root, info, **kwargs):

        pk_dict = {}
        for pk_name in pk_names:
            # Get the new instance data from kwargs[param_name]
            pk_dict[pk_name] = kwargs.get(pk_name)

        with scoped_db_session_from_sa_connection_string(sa_connection_string) as s:
            s: Session

            ###################
            # DELETE DATA
            ###################

            # Find the object and flag it for deletion in the next commit.
            deleted_base_query = s.query(sa_model_class)
            for pk_name, pk_id in pk_dict.items():
                deleted_base_query = deleted_base_query.filter(
                    getattr(sa_model_class, pk_name) == int(pk_id)
                )

            # Delete and return how many objects were deleted.
            try:
                deleted_count = deleted_base_query.delete()
                s.commit()
            except DBAPIError as e:
                s.rollback()
                s.flush()
                s.close()
                raise GraphQLError(e.args)

        # Return the count of deleted objects as 'result'
        partial_delete_obj_class_invocation = {
            deleted_count_arg_name: int(deleted_count)
        }

        #gc.collect()

        return delete_obj_partial_class(**partial_delete_obj_class_invocation)

    definitive_delete_func = mutate_func
    delete_hook = hooks.get(HookOperation.DELETE)
    if delete_hook:
        definitive_delete_func = delete_hook(mutate_func)

    # Create Definitive Class
    definitive_delete_obj_class_items = {"mutate": definitive_delete_func}

    definitive_create_obj_class = type(
        f"Delete{cls_name}",
        (delete_obj_partial_class,),
        definitive_delete_obj_class_items,
    )

    return definitive_create_obj_class
